@@ -32,8 +32,8 @@
  *
  * Spice supports sending keyboard key events and keyboard leds
  * synchronization. The key events are sent using
- * spice_inputs_key_press() and spice_inputs_key_release() using PC AT
- * scancode.
+ * spice_inputs_key_press() and spice_inputs_key_release() using
+ * a modified variant of PC XT scancodes.
  *
  * Guest keyboard leds state can be manipulated with
  * spice_inputs_set_key_locks(). When key lock change, a notification
@@ -147,24 +147,6 @@ static void spice_inputs_channel_class_init(SpiceInputsChannelClass *klass)
     channel_set_handlers(SPICE_CHANNEL_CLASS(klass));
 }
 
-/* signal trampoline---------------------------------------------------------- */
-
-struct SPICE_INPUTS_MODIFIERS {
-};
-
-/* main context */
-static void do_emit_main_context(GObject *object, int signum, gpointer params)
-{
-    switch (signum) {
-    case SPICE_INPUTS_MODIFIERS: {
-        g_signal_emit(object, signals[signum], 0);
-        break;
-    }
-    default:
-        g_warn_if_reached();
-    }
-}
-
 /* ------------------------------------------------------------------ */
 
 static SpiceMsgOut* mouse_motion(SpiceInputsChannel *channel)
@@ -251,7 +233,7 @@ static void inputs_handle_init(SpiceChannel *channel, SpiceMsgIn *in)
     SpiceMsgInputsInit *init = spice_msg_in_parsed(in);
 
     c->modifiers = init->keyboard_modifiers;
-    emit_main_context(channel, SPICE_INPUTS_MODIFIERS);
+    g_coroutine_signal_emit(channel, signals[SPICE_INPUTS_MODIFIERS], 0);
 }
 
 /* coroutine context */
@@ -261,7 +243,7 @@ static void inputs_handle_modifiers(SpiceChannel *channel, SpiceMsgIn *in)
     SpiceMsgInputsKeyModifiers *modifiers = spice_msg_in_parsed(in);
 
     c->modifiers = modifiers->modifiers;
-    emit_main_context(channel, SPICE_INPUTS_MODIFIERS);
+    g_coroutine_signal_emit(channel, signals[SPICE_INPUTS_MODIFIERS], 0);
 }
 
 /* coroutine context */
@@ -456,7 +438,8 @@ void spice_inputs_button_release(SpiceInputsChannel *channel, gint button,
 /**
  * spice_inputs_key_press:
  * @channel:
- * @scancode: a PC AT key scancode
+ * @scancode: a PC XT (set 1) key scancode.  For scancodes with an %0xe0
+ *            prefix, drop the prefix and OR the scancode with %0x100.
  *
  * Press a key.
  **/
@@ -481,7 +464,8 @@ void spice_inputs_key_press(SpiceInputsChannel *channel, guint scancode)
 /**
  * spice_inputs_key_release:
  * @channel:
- * @scancode: a PC AT key scancode
+ * @scancode: a PC XT (set 1) key scancode.  For scancodes with an %0xe0
+ *            prefix, drop the prefix and OR the scancode with %0x100.
  *
  * Release a key.
  **/
@@ -506,7 +490,8 @@ void spice_inputs_key_release(SpiceInputsChannel *channel, guint scancode)
 /**
  * spice_inputs_key_press_and_release:
  * @channel:
- * @scancode: a PC AT key scancode
+ * @scancode: a PC XT (set 1) key scancode.  For scancodes with an %0xe0
+ *            prefix, drop the prefix and OR the scancode with %0x100.
  *
  * Press and release a key event atomically (in the same message).
  *
