@@ -18,9 +18,7 @@
 #ifndef __SPICE_CLIENT_CHANNEL_PRIV_H__
 #define __SPICE_CLIENT_CHANNEL_PRIV_H__
 
-#ifdef HAVE_CONFIG_H
 #include "config.h"
-#endif
 
 #include <openssl/ssl.h>
 #include <gio/gio.h>
@@ -68,6 +66,7 @@ struct _SpiceMsgIn {
 
 enum spice_channel_state {
     SPICE_CHANNEL_STATE_UNCONNECTED = 0,
+    SPICE_CHANNEL_STATE_RECONNECTING,
     SPICE_CHANNEL_STATE_CONNECTING,
     SPICE_CHANNEL_STATE_READY,
     SPICE_CHANNEL_STATE_SWITCHING,
@@ -110,6 +109,8 @@ struct _SpiceChannelPrivate {
 
     char                        name[16];
     enum spice_channel_state    state;
+    SpiceChannelEvent           event;
+
     spice_parse_channel_func_t  parser;
     SpiceMessageMarshallers     *marshallers;
     guint                       channel_watch;
@@ -136,6 +137,7 @@ struct _SpiceChannelPrivate {
     GSList                      *flushing;
 
     gboolean                    disable_channel_msg;
+    gboolean                    auth_needs_username_and_password;
     GError                      *error;
 };
 
@@ -186,17 +188,6 @@ void spice_caps_set(GArray *caps, guint32 cap, const gchar *desc);
 #define spice_channel_set_capability(channel, cap)                      \
     spice_caps_set(SPICE_CHANNEL(channel)->priv->caps, cap, #cap)
 
-/* coroutine context */
-#define emit_main_context(object, event, args...)                                      \
-    G_STMT_START {                                                                     \
-        if (coroutine_self_is_main()) {                                 \
-            do_emit_main_context(G_OBJECT(object), event, &((struct event) { args })); \
-        } else {                                                                       \
-            g_signal_emit_main_context(G_OBJECT(object), do_emit_main_context,         \
-                                       event, &((struct event) { args }), G_STRLOC);   \
-        }                                                                              \
-    } G_STMT_END
-
 gchar *spice_channel_supported_string(void);
 
 void spice_vmc_write_async(SpiceChannel *self,
