@@ -15,15 +15,12 @@
    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, see <http://www.gnu.org/licenses/>.
 */
-#ifdef HAVE_CONFIG_H
 #include "config.h"
-#endif
 
 #include <gst/gst.h>
 #include <gst/app/gstappsrc.h>
-#include <gst/app/gstappbuffer.h>
 #include <gst/app/gstappsink.h>
-#include <gst/interfaces/streamvolume.h>
+#include <gst/audio/streamvolume.h>
 
 #include "spice-gstaudio.h"
 #include "spice-common.h"
@@ -51,9 +48,8 @@ struct _SpiceGstaudioPrivate {
     guint                   mmtime_id;
 };
 
-static void channel_event(SpiceChannel *channel, SpiceChannelEvent event,
-                          gpointer data);
 static gboolean connect_channel(SpiceAudio *audio, SpiceChannel *channel);
+static void channel_weak_notified(gpointer data, GObject *where_the_object_was);
 
 static void spice_gstaudio_finalize(GObject *obj)
 {
@@ -81,34 +77,29 @@ void stream_dispose(struct stream *s)
 
 static void spice_gstaudio_dispose(GObject *obj)
 {
+    SpiceGstaudio *gstaudio = SPICE_GSTAUDIO(obj);
     SpiceGstaudioPrivate *p;
     SPICE_DEBUG("%s", __FUNCTION__);
-    p = SPICE_GSTAUDIO_GET_PRIVATE(obj);
+    p = gstaudio->priv;
 
     stream_dispose(&p->playback);
     stream_dispose(&p->record);
 
-    if (p->pchannel != NULL) {
-        g_signal_handlers_disconnect_by_func(p->pchannel,
-                                             channel_event, obj);
-        g_object_unref(p->pchannel);
-        p->pchannel = NULL;
-    }
+    if (p->pchannel)
+        g_object_weak_unref(G_OBJECT(p->pchannel), channel_weak_notified, gstaudio);
+    p->pchannel = NULL;
 
-    if (p->rchannel != NULL) {
-        g_signal_handlers_disconnect_by_func(p->rchannel,
-                                             channel_event, obj);
-        g_object_unref(p->rchannel);
-        p->rchannel = NULL;
-    }
+    if (p->rchannel)
+        g_object_weak_unref(G_OBJECT(p->rchannel), channel_weak_notified, gstaudio);
+    p->rchannel = NULL;
 
     if (G_OBJECT_CLASS(spice_gstaudio_parent_class)->dispose)
         G_OBJECT_CLASS(spice_gstaudio_parent_class)->dispose(obj);
 }
 
-static void spice_gstaudio_init(SpiceGstaudio *pulse)
+static void spice_gstaudio_init(SpiceGstaudio *gstaudio)
 {
-    pulse->priv = SPICE_GSTAUDIO_GET_PRIVATE(pulse);
+    gstaudio->priv = SPICE_GSTAUDIO_GET_PRIVATE(gstaudio);
 }
 
 static void spice_gstaudio_class_init(SpiceGstaudioClass *klass)
@@ -124,21 +115,22 @@ static void spice_gstaudio_class_init(SpiceGstaudioClass *klass)
     g_type_class_add_private(klass, sizeof(SpiceGstaudioPrivate));
 }
 
-static void record_new_buffer(GstAppSink *appsink, gpointer data)
+static GstFlowReturn record_new_buffer(GstAppSink *appsink, gpointer data)
 {
     SpiceGstaudio *gstaudio = data;
     SpiceGstaudioPrivate *p = gstaudio->priv;
     GstMessage *msg;
 
-    g_return_if_fail(p != NULL);
+    g_return_val_if_fail(p != NULL, GST_FLOW_ERROR);
 
-    msg = gst_message_new_application(GST_OBJECT(p->record.pipe), NULL);
+    msg = gst_message_new_application(GST_OBJECT(p->record.pipe),
+                                      gst_structure_new_empty ("new-sample"));
     gst_element_post_message(p->record.pipe, msg);
+    return GST_FLOW_OK;
 }
 
-static void record_stop(SpiceRecordChannel *channel, gpointer data)
+static void record_stop(SpiceGstaudio *gstaudio)
 {
-    SpiceGstaudio *gstaudio = data;
     SpiceGstaudioPrivate *p = gstaudio->priv;
 
     SPICE_DEBUG("%s", __FUNCTION__);
@@ -155,19 +147,33 @@ static gboolean record_bus_cb(GstBus *bus, GstMessage *msg, gpointer data)
 
     switch (GST_MESSAGE_TYPE(msg)) {
     case GST_MESSAGE_APPLICATION: {
-        GstBuffer *b;
+        GstSample *s;
+        GstBuffer *buffer;
+        GstMapInfo mapping;
 
-        b = gst_app_sink_pull_buffer(GST_APP_SINK(p->record.sink));
-        if (!b) {
+        s = gst_app_sink_pull_sample(GST_APP_SINK(p->record.sink));
+        if (!s) {
+            if (!gst_app_sink_is_eos(GST_APP_SINK(p->record.sink)))
+                g_warning("eos not reached, but can't pull new sample");
+            return TRUE;
+        }
+
+        buffer = gst_sample_get_buffer(s);
+        if (!buffer) {
             if (!gst_app_sink_is_eos(GST_APP_SINK(p->record.sink)))
                 g_warning("eos not reached, but can't pull new buffer");
             return TRUE;
         }
+        if (!gst_buffer_map(buffer, &mapping, GST_MAP_READ)) {
+            return TRUE;
+        }
 
         spice_record_send_data(SPICE_RECORD_CHANNEL(p->rchannel),
                                /* FIXME: server side doesn't care about ts?
                                   what is the unit? ms apparently */
-                               GST_BUFFER_DATA(b), GST_BUFFER_SIZE(b), 0);
+                               mapping.data, mapping.size, 0);
+        gst_buffer_unmap(buffer, &mapping);
+        gst_sample_unref(s);
         break;
     }
     default:
@@ -189,7 +195,7 @@ static void record_start(SpiceRecordChannel *channel, gint format, gint channels
     if (p->record.pipe &&
         (p->record.rate != frequency ||
          p->record.channels != channels)) {
-        record_stop(channel, data);
+        record_stop(gstaudio);
         gst_object_unref(p->record.pipe);
         p->record.pipe = NULL;
     }
@@ -198,16 +204,16 @@ static void record_start(SpiceRecordChannel *channel, gint format, gint channels
         GError *error = NULL;
         GstBus *bus;
         gchar *audio_caps =
-            g_strdup_printf("audio/x-raw-int,channels=%d,rate=%d,signed=(boolean)true,"
-                            "width=16,depth=16,endianness=1234", channels, frequency);
+            g_strdup_printf("audio/x-raw,format=\"S16LE\",channels=%d,rate=%d,"
+                            "layout=interleaved", channels, frequency);
         gchar *pipeline =
             g_strdup_printf("autoaudiosrc name=audiosrc ! queue ! audioconvert ! audioresample ! "
                             "appsink caps=\"%s\" name=appsink", audio_caps);
 
         p->record.pipe = gst_parse_launch(pipeline, &error);
-        if (p->record.pipe == NULL) {
+        if (error != NULL) {
             g_warning("Failed to create pipeline: %s", error->message);
-            goto lerr;
+            goto cleanup;
         }
 
         bus = gst_pipeline_get_bus(GST_PIPELINE(p->record.pipe));
@@ -220,10 +226,14 @@ static void record_start(SpiceRecordChannel *channel, gint format, gint channels
         p->record.channels = channels;
 
         gst_app_sink_set_emit_signals(GST_APP_SINK(p->record.sink), TRUE);
-        spice_g_signal_connect_object(p->record.sink, "new-buffer",
+        spice_g_signal_connect_object(p->record.sink, "new-sample",
                                       G_CALLBACK(record_new_buffer), gstaudio, 0);
 
-lerr:
+cleanup:
+        if (error != NULL && p->record.pipe != NULL) {
+            gst_object_unref(p->record.pipe);
+            p->record.pipe = NULL;
+        }
         g_clear_error(&error);
         g_free(audio_caps);
         g_free(pipeline);
@@ -233,35 +243,10 @@ lerr:
         gst_element_set_state(p->record.pipe, GST_STATE_PLAYING);
 }
 
-static void channel_event(SpiceChannel *channel, SpiceChannelEvent event,
-                          gpointer data)
-{
-    SpiceGstaudio *gstaudio = data;
-    SpiceGstaudioPrivate *p = gstaudio->priv;
-
-    switch (event) {
-    case SPICE_CHANNEL_OPENED:
-        break;
-    case SPICE_CHANNEL_CLOSED:
-        if (channel == p->pchannel) {
-            p->pchannel = NULL;
-            g_object_unref(channel);
-        } else if (channel == p->rchannel) {
-            record_stop(SPICE_RECORD_CHANNEL(channel), gstaudio);
-            p->rchannel = NULL;
-            g_object_unref(channel);
-        } else /* if (p->pchannel || p->rchannel) */
-            g_warn_if_reached();
-        break;
-    default:
-        break;
-    }
-}
-
 static void playback_stop(SpicePlaybackChannel *channel, gpointer data)
 {
     SpiceGstaudio *gstaudio = data;
-    SpiceGstaudioPrivate *p = SPICE_GSTAUDIO_GET_PRIVATE(gstaudio);
+    SpiceGstaudioPrivate *p = gstaudio->priv;
 
     if (p->playback.pipe)
         gst_element_set_state(p->playback.pipe, GST_STATE_READY);
@@ -274,7 +259,7 @@ static void playback_stop(SpicePlaybackChannel *channel, gpointer data)
 static gboolean update_mmtime_timeout_cb(gpointer data)
 {
     SpiceGstaudio *gstaudio = data;
-    SpiceGstaudioPrivate *p = SPICE_GSTAUDIO_GET_PRIVATE(gstaudio);
+    SpiceGstaudioPrivate *p = gstaudio->priv;
     GstQuery *q;
 
     q = gst_query_new_latency();
@@ -296,7 +281,7 @@ static void playback_start(SpicePlaybackChannel *channel, gint format, gint chan
                            gint frequency, gpointer data)
 {
     SpiceGstaudio *gstaudio = data;
-    SpiceGstaudioPrivate *p = SPICE_GSTAUDIO_GET_PRIVATE(gstaudio);
+    SpiceGstaudioPrivate *p = gstaudio->priv;
 
     g_return_if_fail(p != NULL);
     g_return_if_fail(format == SPICE_AUDIO_FMT_S16);
@@ -312,24 +297,28 @@ static void playback_start(SpicePlaybackChannel *channel, gint format, gint chan
     if (!p->playback.pipe) {
         GError *error = NULL;
         gchar *audio_caps =
-            g_strdup_printf("audio/x-raw-int,channels=%d,rate=%d,signed=(boolean)true,"
-                            "width=16,depth=16,endianness=1234", channels, frequency);
+            g_strdup_printf("audio/x-raw,format=\"S16LE\",channels=%d,rate=%d,"
+                            "layout=interleaved", channels, frequency);
         gchar *pipeline = g_strdup (g_getenv("SPICE_GST_AUDIOSINK"));
         if (pipeline == NULL)
             pipeline = g_strdup_printf("appsrc is-live=1 do-timestamp=0 caps=\"%s\" name=\"appsrc\" ! queue ! "
                                        "audioconvert ! audioresample ! autoaudiosink name=\"audiosink\"", audio_caps);
         SPICE_DEBUG("audio pipeline: %s", pipeline);
         p->playback.pipe = gst_parse_launch(pipeline, &error);
-        if (p->playback.pipe == NULL) {
+        if (error != NULL) {
             g_warning("Failed to create pipeline: %s", error->message);
-            goto lerr;
+            goto cleanup;
         }
         p->playback.src = gst_bin_get_by_name(GST_BIN(p->playback.pipe), "appsrc");
         p->playback.sink = gst_bin_get_by_name(GST_BIN(p->playback.pipe), "audiosink");
         p->playback.rate = frequency;
         p->playback.channels = channels;
 
-lerr:
+cleanup:
+        if (error != NULL && p->playback.pipe != NULL) {
+            gst_object_unref(p->playback.pipe);
+            p->playback.pipe = NULL;
+        }
         g_clear_error(&error);
         g_free(audio_caps);
         g_free(pipeline);
@@ -349,13 +338,13 @@ static void playback_data(SpicePlaybackChannel *channel,
                           gpointer data)
 {
     SpiceGstaudio *gstaudio = data;
-    SpiceGstaudioPrivate *p = SPICE_GSTAUDIO_GET_PRIVATE(gstaudio);
+    SpiceGstaudioPrivate *p = gstaudio->priv;
     GstBuffer *buf;
 
     g_return_if_fail(p != NULL);
 
     audio = g_memdup(audio, size); /* TODO: try to avoid memory copy */
-    buf = gst_app_buffer_new(audio, size, g_free, audio);
+    buf = gst_buffer_new_wrapped(audio, size);
     gst_app_src_push_buffer(GST_APP_SRC(p->playback.src), buf);
 }
 
@@ -481,6 +470,22 @@ static void record_mute_changed(GObject *object, GParamSpec *pspec, gpointer dat
     g_object_unref(e);
 }
 
+static void
+channel_weak_notified(gpointer data,
+                      GObject *where_the_object_was)
+{
+    SpiceGstaudio *gstaudio = SPICE_GSTAUDIO(data);
+    SpiceGstaudioPrivate *p = gstaudio->priv;
+
+    if (where_the_object_was == (GObject *)p->pchannel) {
+        p->pchannel = NULL;
+    } else if (where_the_object_was == (GObject *)p->rchannel) {
+        SPICE_DEBUG("record closed");
+        record_stop(gstaudio);
+        p->rchannel = NULL;
+    }
+}
+
 static gboolean connect_channel(SpiceAudio *audio, SpiceChannel *channel)
 {
     SpiceGstaudio *gstaudio = SPICE_GSTAUDIO(audio);
@@ -489,15 +494,14 @@ static gboolean connect_channel(SpiceAudio *audio, SpiceChannel *channel)
     if (SPICE_IS_PLAYBACK_CHANNEL(channel)) {
         g_return_val_if_fail(p->pchannel == NULL, FALSE);
 
-        p->pchannel = g_object_ref(channel);
+        p->pchannel = channel;
+        g_object_weak_ref(G_OBJECT(p->pchannel), channel_weak_notified, audio);
         spice_g_signal_connect_object(channel, "playback-start",
                                       G_CALLBACK(playback_start), gstaudio, 0);
         spice_g_signal_connect_object(channel, "playback-data",
                                       G_CALLBACK(playback_data), gstaudio, 0);
         spice_g_signal_connect_object(channel, "playback-stop",
                                       G_CALLBACK(playback_stop), gstaudio, 0);
-        spice_g_signal_connect_object(channel, "channel-event",
-                                      G_CALLBACK(channel_event), gstaudio, 0);
         spice_g_signal_connect_object(channel, "notify::volume",
                                       G_CALLBACK(playback_volume_changed), gstaudio, 0);
         spice_g_signal_connect_object(channel, "notify::mute",
@@ -509,13 +513,12 @@ static gboolean connect_channel(SpiceAudio *audio, SpiceChannel *channel)
     if (SPICE_IS_RECORD_CHANNEL(channel)) {
         g_return_val_if_fail(p->rchannel == NULL, FALSE);
 
-        p->rchannel = g_object_ref(channel);
+        p->rchannel = channel;
+        g_object_weak_ref(G_OBJECT(p->rchannel), channel_weak_notified, audio);
         spice_g_signal_connect_object(channel, "record-start",
                                       G_CALLBACK(record_start), gstaudio, 0);
         spice_g_signal_connect_object(channel, "record-stop",
-                                      G_CALLBACK(record_stop), gstaudio, 0);
-        spice_g_signal_connect_object(channel, "channel-event",
-                                      G_CALLBACK(channel_event), gstaudio, 0);
+                                      G_CALLBACK(record_stop), gstaudio, G_CONNECT_SWAPPED);
         spice_g_signal_connect_object(channel, "notify::volume",
                                       G_CALLBACK(record_volume_changed), gstaudio, 0);
         spice_g_signal_connect_object(channel, "notify::mute",
