@@ -74,16 +74,16 @@ static const char *context_state_names[] = {
 #define STATE_NAME(array, state) \
     ((state < G_N_ELEMENTS(array)) ? array[state] : NULL)
 
-static void channel_event(SpiceChannel *channel, SpiceChannelEvent event,
-                          gpointer data);
 static void stream_stop(SpicePulse *pulse, struct stream *s);
 static gboolean connect_channel(SpiceAudio *audio, SpiceChannel *channel);
+static void channel_weak_notified(gpointer data, GObject *where_the_object_was);
 
 static void spice_pulse_finalize(GObject *obj)
 {
+    SpicePulse *pulse = SPICE_PULSE(obj);
     SpicePulsePrivate *p;
 
-    p = SPICE_PULSE_GET_PRIVATE(obj);
+    p = pulse->priv;
 
     if (p->context != NULL)
         pa_context_unref(p->context);
@@ -96,10 +96,11 @@ static void spice_pulse_finalize(GObject *obj)
 
 static void spice_pulse_dispose(GObject *obj)
 {
+    SpicePulse *pulse = SPICE_PULSE(obj);
     SpicePulsePrivate *p;
 
     SPICE_DEBUG("%s", __FUNCTION__);
-    p = SPICE_PULSE_GET_PRIVATE(obj);
+    p = pulse->priv;
 
     if (p->playback.uncork_op)
         pa_operation_unref(p->playback.uncork_op);
@@ -118,11 +119,11 @@ static void spice_pulse_dispose(GObject *obj)
     p->record.cork_op = NULL;
 
     if (p->pchannel)
-        g_object_unref(p->pchannel);
+        g_object_weak_unref(G_OBJECT(p->pchannel), channel_weak_notified, pulse);
     p->pchannel = NULL;
 
     if (p->rchannel)
-        g_object_unref(p->rchannel);
+        g_object_weak_unref(G_OBJECT(p->rchannel), channel_weak_notified, pulse);
     p->rchannel = NULL;
 
     G_OBJECT_CLASS(spice_pulse_parent_class)->dispose(obj);
@@ -160,7 +161,7 @@ static void pulse_uncork_cb(pa_stream *pastream, int success, void *data)
 
 static void stream_uncork(SpicePulse *pulse, struct stream *s)
 {
-    SpicePulsePrivate *p = SPICE_PULSE_GET_PRIVATE(pulse);
+    SpicePulsePrivate *p = pulse->priv;
     pa_operation *o = NULL;
 
     g_return_if_fail(s->stream);
@@ -219,7 +220,7 @@ static void pulse_cork_cb(pa_stream *pastream, int success, void *data)
 
 static void stream_cork(SpicePulse *pulse, struct stream *s, gboolean with_flush)
 {
-    SpicePulsePrivate *p = SPICE_PULSE_GET_PRIVATE(pulse);
+    SpicePulsePrivate *p = pulse->priv;
     pa_operation *o = NULL;
 
     if (s->uncork_op) {
@@ -242,7 +243,7 @@ static void stream_cork(SpicePulse *pulse, struct stream *s, gboolean with_flush
 
 static void stream_stop(SpicePulse *pulse, struct stream *s)
 {
-    SpicePulsePrivate *p = SPICE_PULSE_GET_PRIVATE(pulse);
+    SpicePulsePrivate *p = pulse->priv;
 
     if (pa_stream_disconnect(s->stream) < 0) {
         g_warning("pa_stream_disconnect() failed: %s",
@@ -254,8 +255,10 @@ static void stream_stop(SpicePulse *pulse, struct stream *s)
 
 static void stream_state_callback(pa_stream *s, void *userdata)
 {
+    SpicePulse *pulse = userdata;
     SpicePulsePrivate *p;
-    p = SPICE_PULSE_GET_PRIVATE(userdata);
+
+    p = pulse->priv;
 
     g_return_if_fail(p != NULL);
     g_return_if_fail(s != NULL);
@@ -273,11 +276,12 @@ static void stream_state_callback(pa_stream *s, void *userdata)
 
 static void stream_underflow_cb(pa_stream *s, void *userdata)
 {
+    SpicePulse *pulse = userdata;
     SpicePulsePrivate *p;
 
     SPICE_DEBUG("PA stream underflow!!");
 
-    p = SPICE_PULSE_GET_PRIVATE(userdata);
+    p = pulse->priv;
     g_return_if_fail(p != NULL);
     p->playback.num_underflow++;
 #ifdef PULSE_ADJUST_LATENCY
@@ -303,7 +307,7 @@ static void stream_update_latency_callback(pa_stream *s, void *userdata)
     int negative = 0;
     SpicePulsePrivate *p;
 
-    p = SPICE_PULSE_GET_PRIVATE(pulse);
+    p = pulse->priv;
 
     g_return_if_fail(s != NULL);
     g_return_if_fail(p != NULL);
@@ -331,7 +335,7 @@ static void stream_update_latency_callback(pa_stream *s, void *userdata)
 
 static void create_playback(SpicePulse *pulse)
 {
-    SpicePulsePrivate *p = SPICE_PULSE_GET_PRIVATE(pulse);
+    SpicePulsePrivate *p = pulse->priv;
     pa_stream_flags_t flags;
     pa_buffer_attr buffer_attr = { 0, };
 
@@ -364,7 +368,7 @@ static void playback_start(SpicePlaybackChannel *channel, gint format, gint chan
                            gint frequency, gpointer data)
 {
     SpicePulse *pulse = data;
-    SpicePulsePrivate *p = SPICE_PULSE_GET_PRIVATE(pulse);
+    SpicePulsePrivate *p = pulse->priv;
     pa_context_state_t state;
     guint latency;
 
@@ -414,7 +418,7 @@ static void playback_data(SpicePlaybackChannel *channel,
                           gpointer data)
 {
     SpicePulse *pulse = data;
-    SpicePulsePrivate *p = SPICE_PULSE_GET_PRIVATE(pulse);
+    SpicePulsePrivate *p = pulse->priv;
     pa_stream_state_t state;
 
     if (!p->playback.stream)
@@ -461,7 +465,7 @@ static void playback_stop(SpicePlaybackChannel *channel, gpointer data)
 static void stream_read_callback(pa_stream *s, size_t length, void *data)
 {
     SpicePulse *pulse = data;
-    SpicePulsePrivate *p = SPICE_PULSE_GET_PRIVATE(pulse);
+    SpicePulsePrivate *p = pulse->priv;
 
     g_return_if_fail(p != NULL);
 
@@ -493,7 +497,7 @@ static void stream_read_callback(pa_stream *s, size_t length, void *data)
 
 static void create_record(SpicePulse *pulse)
 {
-    SpicePulsePrivate *p = SPICE_PULSE_GET_PRIVATE(pulse);
+    SpicePulsePrivate *p = pulse->priv;
     pa_buffer_attr buffer_attr = { 0, };
     pa_stream_flags_t flags;
 
@@ -525,7 +529,7 @@ static void record_start(SpiceRecordChannel *channel, gint format, gint channels
                          gint frequency, gpointer data)
 {
     SpicePulse *pulse = data;
-    SpicePulsePrivate *p = SPICE_PULSE_GET_PRIVATE(pulse);
+    SpicePulsePrivate *p = pulse->priv;
     pa_context_state_t state;
 
     p->record.started = TRUE;
@@ -562,9 +566,8 @@ static void record_start(SpiceRecordChannel *channel, gint format, gint channels
     p->state = state;
 }
 
-static void record_stop(SpiceRecordChannel *channel, gpointer data)
+static void record_stop(SpicePulse *pulse)
 {
-    SpicePulse *pulse = data;
     SpicePulsePrivate *p = pulse->priv;
 
     SPICE_DEBUG("%s", __FUNCTION__);
@@ -576,33 +579,6 @@ static void record_stop(SpiceRecordChannel *channel, gpointer data)
     stream_stop(pulse, &p->record);
 }
 
-static void channel_event(SpiceChannel *channel, SpiceChannelEvent event,
-                          gpointer data)
-{
-    SpicePulse *pulse = data;
-    SpicePulsePrivate *p = pulse->priv;
-
-    switch (event) {
-    case SPICE_CHANNEL_OPENED:
-        break;
-    case SPICE_CHANNEL_CLOSED:
-        if (channel == p->pchannel) {
-            SPICE_DEBUG("playback closed");
-            p->pchannel = NULL;
-            g_object_unref(channel);
-        } else if (channel == p->rchannel) {
-            SPICE_DEBUG("record closed");
-            record_stop(SPICE_RECORD_CHANNEL(channel), pulse);
-            p->rchannel = NULL;
-            g_object_unref(channel);
-        } else /* if (p->pchannel || p->rchannel) */
-            g_warn_if_reached();
-        break;
-    default:
-        break;
-    }
-}
-
 static void playback_volume_changed(GObject *object, GParamSpec *pspec, gpointer data)
 {
     SpicePulse *pulse = data;
@@ -743,6 +719,22 @@ static void record_volume_changed(GObject *object, GParamSpec *pspec, gpointer d
         pa_operation_unref(op);
 }
 
+static void
+channel_weak_notified(gpointer data,
+                      GObject *where_the_object_was)
+{
+    SpicePulse *pulse = SPICE_PULSE(data);
+    SpicePulsePrivate *p = pulse->priv;
+
+    if (where_the_object_was == (GObject *)p->pchannel) {
+        p->pchannel = NULL;
+    } else if (where_the_object_was == (GObject *)p->rchannel) {
+        SPICE_DEBUG("record closed");
+        record_stop(pulse);
+        p->rchannel = NULL;
+    }
+}
+
 static gboolean connect_channel(SpiceAudio *audio, SpiceChannel *channel)
 {
     SpicePulse *pulse = SPICE_PULSE(audio);
@@ -751,15 +743,14 @@ static gboolean connect_channel(SpiceAudio *audio, SpiceChannel *channel)
     if (SPICE_IS_PLAYBACK_CHANNEL(channel)) {
         g_return_val_if_fail(p->pchannel == NULL, FALSE);
 
-        p->pchannel = g_object_ref(channel);
+        p->pchannel = channel;
+        g_object_weak_ref(G_OBJECT(p->pchannel), channel_weak_notified, audio);
         spice_g_signal_connect_object(channel, "playback-start",
                                       G_CALLBACK(playback_start), pulse, 0);
         spice_g_signal_connect_object(channel, "playback-data",
                                       G_CALLBACK(playback_data), pulse, 0);
         spice_g_signal_connect_object(channel, "playback-stop",
                                       G_CALLBACK(playback_stop), pulse, 0);
-        spice_g_signal_connect_object(channel, "channel-event",
-                                      G_CALLBACK(channel_event), pulse, 0);
         spice_g_signal_connect_object(channel, "notify::volume",
                                       G_CALLBACK(playback_volume_changed), pulse, 0);
         spice_g_signal_connect_object(channel, "notify::mute",
@@ -773,13 +764,12 @@ static gboolean connect_channel(SpiceAudio *audio, SpiceChannel *channel)
     if (SPICE_IS_RECORD_CHANNEL(channel)) {
         g_return_val_if_fail(p->rchannel == NULL, FALSE);
 
-        p->rchannel = g_object_ref(channel);
+        p->rchannel = channel;
+        g_object_weak_ref(G_OBJECT(p->rchannel), channel_weak_notified, audio);
         spice_g_signal_connect_object(channel, "record-start",
                                       G_CALLBACK(record_start), pulse, 0);
         spice_g_signal_connect_object(channel, "record-stop",
-                                      G_CALLBACK(record_stop), pulse, 0);
-        spice_g_signal_connect_object(channel, "channel-event",
-                                      G_CALLBACK(channel_event), pulse, 0);
+                                      G_CALLBACK(record_stop), pulse, G_CONNECT_SWAPPED);
         spice_g_signal_connect_object(channel, "notify::volume",
                                       G_CALLBACK(record_volume_changed), pulse, 0);
         spice_g_signal_connect_object(channel, "notify::mute",
@@ -793,8 +783,10 @@ static gboolean connect_channel(SpiceAudio *audio, SpiceChannel *channel)
 
 static void context_state_callback(pa_context *c, void *userdata)
 {
+    SpicePulse *pulse = userdata;
     SpicePulsePrivate *p;
-    p = SPICE_PULSE_GET_PRIVATE(userdata);
+
+    p = pulse->priv;
 
     g_return_if_fail(p != NULL);
     g_return_if_fail(c != NULL);
@@ -836,7 +828,7 @@ SpicePulse *spice_pulse_new(SpiceSession *session, GMainContext *context,
                          "session", session,
                          "main-context", context,
                          NULL);
-    p = SPICE_PULSE_GET_PRIVATE(pulse);
+    p = pulse->priv;
 
     p->mainloop = pa_glib_mainloop_new(context);
     p->state = PA_CONTEXT_READY;
