@@ -16,9 +16,7 @@
    License along with this library; if not, see <http://www.gnu.org/licenses/>.
 */
 
-#ifdef HAVE_CONFIG_H
-# include "config.h"
-#endif
+#include "config.h"
 #include <glib/gi18n.h>
 
 #include <sys/stat.h>
@@ -73,7 +71,7 @@ struct _SpiceWindow {
     bool             fullscreen;
     bool             mouse_grabbed;
     SpiceChannel     *display_channel;
-#ifdef WIN32
+#ifdef G_OS_WIN32
     gint             win_x;
     gint             win_y;
 #endif
@@ -185,7 +183,7 @@ static struct {
     { .text = N_("TLS Port"),   .prop = "tls-port"  },
 };
 
-#ifndef WIN32
+#ifndef G_OS_WIN32
 static void recent_selection_changed_dialog_cb(GtkRecentChooser *chooser, gpointer data)
 {
     GtkRecentInfo *info;
@@ -264,7 +262,7 @@ static int connect_dialog(SpiceSession *session)
     label = gtk_label_new("Recent connections:");
     gtk_box_pack_start(GTK_BOX(area), label, TRUE, TRUE, 0);
     gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
-#ifndef WIN32
+#ifndef G_OS_WIN32
     GtkRecentFilter *rfilter;
     GtkWidget *recent;
 
@@ -397,13 +395,13 @@ static void menu_cb_paste(GtkAction *action, void *data)
 static void window_set_fullscreen(SpiceWindow *win, gboolean fs)
 {
     if (fs) {
-#ifdef WIN32
+#ifdef G_OS_WIN32
         gtk_window_get_position(GTK_WINDOW(win->toplevel), &win->win_x, &win->win_y);
 #endif
         gtk_window_fullscreen(GTK_WINDOW(win->toplevel));
     } else {
         gtk_window_unfullscreen(GTK_WINDOW(win->toplevel));
-#ifdef WIN32
+#ifdef G_OS_WIN32
         gtk_window_move(GTK_WINDOW(win->toplevel), win->win_x, win->win_y);
 #endif
     }
@@ -933,7 +931,7 @@ static gboolean is_gtk_session_property(const gchar *property)
     return FALSE;
 }
 
-#ifndef WIN32
+#ifndef G_OS_WIN32
 static void recent_item_activated_cb(GtkRecentChooser *chooser, gpointer data)
 {
     GtkRecentInfo *info;
@@ -1033,7 +1031,7 @@ static SpiceWindow *create_spice_window(spice_connection *conn, SpiceChannel *ch
     win->ritem  = gtk_ui_manager_get_widget
         (win->ui, "/MainMenu/FileMenu/FileRecentMenu");
 
-#ifndef WIN32
+#ifndef G_OS_WIN32
     GtkRecentFilter  *rfilter;
 
     win->rmenu = gtk_recent_chooser_menu_new();
@@ -1195,10 +1193,14 @@ static void recent_add(SpiceSession *session)
     g_object_get(session, "uri", &uri, NULL);
     SPICE_DEBUG("%s: %s", __FUNCTION__, uri);
 
-    g_return_if_fail(g_str_has_prefix(uri, "spice://"));
-
     recent = gtk_recent_manager_get_default();
-    meta.display_name = uri + 8;
+    if (g_str_has_prefix(uri, "spice://"))
+        meta.display_name = uri + 8;
+    else if (g_str_has_prefix(uri, "spice+unix://"))
+        meta.display_name = uri + 13;
+    else
+        g_return_if_reached();
+
     if (!gtk_recent_manager_add_full(recent, uri, &meta))
         g_warning("Recent item couldn't be added successfully");
 
@@ -1701,7 +1703,7 @@ static void (* segv_handler) (int) = SIG_DFL;
 static void (* abrt_handler) (int) = SIG_DFL;
 static void (* fpe_handler)  (int) = SIG_DFL;
 static void (* ill_handler)  (int) = SIG_DFL;
-#ifndef WIN32
+#ifndef G_OS_WIN32
 static void (* bus_handler)  (int) = SIG_DFL;
 #endif
 
@@ -1718,7 +1720,7 @@ signal_handler(int signum)
     signal(SIGABRT, abrt_handler);
     signal(SIGFPE,  fpe_handler);
     signal(SIGILL,  ill_handler);
-#ifndef WIN32
+#ifndef G_OS_WIN32
     signal(SIGBUS,  bus_handler);
 #endif
 
@@ -1790,7 +1792,7 @@ int main(int argc, char *argv[])
     GOptionContext *context;
     spice_connection *conn;
     gchar *conf_file, *conf;
-    char *host = NULL, *port = NULL, *tls_port = NULL;
+    char *host = NULL, *port = NULL, *tls_port = NULL, *unix_path = NULL;
 
 #if !GLIB_CHECK_VERSION(2,31,18)
     g_thread_init(NULL);
@@ -1805,7 +1807,7 @@ int main(int argc, char *argv[])
     abrt_handler = signal(SIGABRT, signal_handler);
     fpe_handler  = signal(SIGFPE,  signal_handler);
     ill_handler  = signal(SIGILL,  signal_handler);
-#ifndef WIN32
+#ifndef G_OS_WIN32
     signal(SIGHUP, signal_handler);
     bus_handler  = signal(SIGBUS,  signal_handler);
 #endif
@@ -1845,7 +1847,9 @@ int main(int argc, char *argv[])
         exit(0);
     }
 
+#if !GLIB_CHECK_VERSION(2,36,0)
     g_type_init();
+#endif
     mainloop = g_main_loop_new(NULL, false);
 
     conn = connection_new();
@@ -1853,13 +1857,14 @@ int main(int argc, char *argv[])
     spice_cmdline_session_setup(conn->session);
 
     g_object_get(conn->session,
+                 "unix-path", &unix_path,
                  "host", &host,
                  "port", &port,
                  "tls-port", &tls_port,
                  NULL);
     /* If user doesn't provide hostname and port, show the dialog window
        instead of connecting to server automatically */
-    if (host == NULL || (port == NULL && tls_port == NULL)) {
+    if ((host == NULL || (port == NULL && tls_port == NULL)) && unix_path == NULL) {
         int ret = connect_dialog(conn->session);
         if (ret != 0) {
             exit(0);
@@ -1868,6 +1873,7 @@ int main(int argc, char *argv[])
     g_free(host);
     g_free(port);
     g_free(tls_port);
+    g_free(unix_path);
 
     watch_stdin();
 
